/*
Copyright 2018 Rene Halver, Forschungszentrum Juelich GmbH, Germany
Copyright 2018 Godehard Sutmann, Forschungszentrum Juelich GmbH, Germany

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may 
   be used to endorse or promote products derived from this software without 
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef ALL_MAIN_HEADER_INC
#define ALL_MAIN_HEADER_INC

#include "ALL_LB.hpp"
#include "ALL_Point.hpp"
#include "ALL_CustomExceptions.hpp"
#include "ALL_Tensor.hpp"
#ifdef ALL_VORONOI
#include "ALL_Voronoi.hpp"
#endif
#include "ALL_Staggered.hpp"
#include "ALL_Unstructured.hpp"
#include "ALL_Histogram.hpp"
#include <vector>
#include <sstream>
#include <string>
#include <algorithm>
#include <numeric>
#include <iomanip>
#include <memory>

#ifdef ALL_VTK_OUTPUT
#include <vtkXMLPUnstructuredGridWriter.h>
#include <vtkXMLUnstructuredGridWriter.h>
#include <vtkUnstructuredGrid.h>
#include <vtkVoxel.h>
#include <vtkPolyhedron.h>
#include <vtkSmartPointer.h>
#include <vtkFloatArray.h>
#include <vtkIntArray.h>
#include <vtkCellData.h>
#include <vtkCellArray.h>
#include <vtkMPIController.h>
#include <vtkProgrammableFilter.h>
#include <vtkInformation.h>
#endif

#define ALL_ESTIMATION_LIMIT_DEFAULT 0

enum ALL_LB_t : int 
            {
                // staggered grid load balancing
                STAGGERED = 0, 
                // tensor based load balancing
                TENSOR = 1, 
                // unstructured-mesh load balancing
                UNSTRUCTURED = 2,
                // voronoi cell based load balancing
                VORONOI = 3,
                // histogram based load balancing
                HISTOGRAM = 4 
            }; 

// T = data type for vertices, W data type for work
template <class T, class W> class ALL
{
    public:

        // constructor (empty)
        ALL() : 
#ifdef ALL_VTK_OUTPUT
                vtk_init(false),
#endif        
                loadbalancing_step(0)
                {
                    // determine correct MPI data type for template T
                    if (std::is_same<T,double>::value) mpi_data_type_T = MPI_DOUBLE;
                    else if (std::is_same<T,float>::value) mpi_data_type_T = MPI_FLOAT;
                    else if (std::is_same<T,int>::value) mpi_data_type_T = MPI_INT;
                    else if (std::is_same<T,long>::value) mpi_data_type_T = MPI_LONG;

                    // determine correct MPI data type for template W 
                    if (std::is_same<W,double>::value) mpi_data_type_W = MPI_DOUBLE;
                    else if (std::is_same<W,float>::value) mpi_data_type_W = MPI_FLOAT;
                    else if (std::is_same<W,int>::value) mpi_data_type_W = MPI_INT;
                    else if (std::is_same<W,long>::value) mpi_data_type_W = MPI_LONG;
                }

        // constructor (dimension only)
        ALL(ALL_LB_t m, const int d, const T g) : ALL()       
        {
            method = m;
            switch(method)
            {
                case ALL_LB_t::TENSOR:
                    balancer.reset(new ALL_Tensor_LB<T,W>(d,(W)0,g));
                    break;
                case ALL_LB_t::STAGGERED:
                    balancer.reset(new ALL_Staggered_LB<T,W>(d,(W)0,g));
                    break;
                case ALL_LB_t::UNSTRUCTURED:
                    balancer.reset(new ALL_Unstructured_LB<T,W>(d,(W)0,g));
                    break;
                case ALL_LB_t::VORONOI:
#ifdef ALL_VORONOI
                    balancer.reset(new ALL_Voronoi_LB<T,W>(d,(W)0,g));
#endif                
                    break;
                case ALL_LB_t::HISTOGRAM:
                    balancer.reset(new ALL_Histogram_LB<T,W>(d,std::vector<W>(10),g));
                    break;
                default:
                    throw ALL_Invalid_Argument_Exception(
                                            __FILE__,
                                            __func__,
                                            __LINE__,
                                            "Unknown type of loadbalancing passed.");
            }
            balancer->set_dimension(d);        
            balancer->set_gamma(g);
            estimation_limit = ALL_ESTIMATION_LIMIT_DEFAULT;
        }

        // constructor (dimension + vertices)
        ALL(ALL_LB_t method, const int d_, std::vector<ALL_Point<T>>& inp, const T g) : ALL(method, d_,g)
        {
            balancer->set_vertices(inp);
            calculate_outline();
        }

        // destructor
        ~ALL() = default;

        // overwrite vertices
        // with a vector
        void set_vertices(std::vector<ALL_Point<T>>&);
        // with an array of T values (unsafe due to missing possibility to check T array)
        void set_vertices(const int n, const int dim, const T*);

        // set up work of domain

        // set up MPI communicator
        void set_communicator(MPI_Comm);
        
        // set gamma (correction value)
        void set_gamma(const T g) { gamma = g;};

        // single value (staggered grid / tensor-based)
        void set_work(const W);

        // vector of work (cell based -> work per cell)
        void set_work(const std::vector<W>&);

        // setup balancer
        void setup();

        // calculate load-balance
        std::vector<ALL_Point<T>>& balance(bool internal = false);

        // separate call ONLY for the setup method of the balancer
        void setup_balancer();

        // getter functions

        // vertices 
        std::vector<ALL_Point<T>>& get_vertices() {return balancer->get_vertices();};
        std::vector<ALL_Point<T>>& get_prev_vertices() {return balancer->get_prev_vertices();};

        // dimension
        int get_dimension() {return balancer->get_dimension();}

        // work
        void get_work(std::vector<W>&);
        void get_work(W&);

        // neighbors
        int get_n_neighbors();
        // provide list of neighbors (ranks)
        void get_neighbors(std::vector<int>&);
        // provide a list of neighbor vertices (VORONOI only)
        void get_neighbor_vertices(std::vector<T>&);
        // provide list of neighbors in each direction
        void get_neighbors(int**);

        // set system size
        // (x_min, x_max, y_min, y_max, z_min, z_max)
        void set_sys_size(std::vector<T>&);

        // method to set method specific data, without
        // the necessity to create dummy hooks for other
        // methods
        void set_method_data(void*);

        // if using a non-cartesian communicator set the processor grid
        // parameters to create an internal cartesian grid
        void set_proc_grid_params(const std::vector<int>&, const std::vector<int>&);
        void set_proc_grid_params(int*, int*);

	// set tag of this process for output
	void set_proc_tag(int);

        void set_min_domain_size(const std::vector<T>&);

#ifdef ALL_VTK_OUTPUT        
        // print outlines and points within domain
        void print_vtk_outlines(int);
        // print vertices 
        void print_vtk_vertices(int);
#endif        

    private:
        ALL_LB_t method;

        // outer cuboid encasing the domain (for cuboid domains identical to domain)
        // defined by front left lower [0][*] and back upper right points [1][*]
        // where * is 0,...,dim-1
        std::vector<ALL_Point<T>> outline;

//         // gamma = correction factor
//         T gamma;
// 
//         // dimension of the system (>=1)
//         int dimension;

        // balancer
        std::unique_ptr<ALL_LB<T,W>> balancer;

        // data containing cells (TODO: defintion of data layout, etc.)
        std::vector<W> work_array;

        // MPI communicator used in library
        MPI_Comm communicator;
        
        // number of vertices (for non-cuboid domains)
        int n_vertices;

        // number of neighbors
        int n_neighbors;

        // calculate the outline of the domain using the vertices
        void calculate_outline();

        // limit for the estimation procedure
        int estimation_limit;

        // position of the local process in the process grid
        std::vector<int> proc_grid_loc;
        // number of processes in each direction
        std::vector<int> proc_grid_dim;
        // check if process grid values are set
        bool proc_grid_set;
	// tag of the process
	int proc_tag;

        // type for MPI communication
        MPI_Datatype mpi_data_type_T;
        MPI_Datatype mpi_data_type_W;

        int loadbalancing_step;

#ifdef ALL_VTK_OUTPUT
        bool vtk_init;
        vtkMPIController* controller;
        vtkMPIController* controller_vertices;
#endif        
};

template <class T, class W> void ALL<T,W>::set_vertices(std::vector<ALL_Point<T>>& inp)
{
    int dim = inp.at(0).get_dimension();
    if (dim != balancer->get_dimension()) throw ALL_Point_Dimension_Missmatch_Exception(
                                __FILE__,
                                __func__,
                                __LINE__,
                                "Dimension of ALL_Points in input vector do not match dimension of ALL object.");
    balancer->set_vertices(inp);
    calculate_outline();
}

template <class T, class W> void ALL<T,W>::set_vertices(const int n, const int dim, const T* inp)
{
    std::vector<ALL_Point<T>> points(n,ALL_Point<T>(dim));
    if (dim != balancer->get_dimension()) throw ALL_Point_Dimension_Missmatch_Exception(
                                __FILE__,
                                __func__,
                                __LINE__,
                                "Dimension of ALL_Points in input vector do not match dimension of ALL object.");
    for (int i = 0; i < n; ++i)
    {
        for (int d = 0; d < dim; ++d)
        {
            points.at(i).set_coordinate(d,inp[ i * dim + d ]);
        }
    }
    balancer->set_vertices(points);
    calculate_outline();
}

template <class T, class W> void ALL<T,W>::set_work(W work)
{
    // set new value for work (single value for whole domain)
    balancer->set_work(work);
}

template <class T, class W> void ALL<T,W>::set_work(const std::vector<W>& work)
{
    balancer->set_work(work);
}

template <class T, class W> void ALL<T,W>::set_proc_grid_params(const std::vector<int>& loc, const std::vector<int>& size)
{
    proc_grid_loc.insert(proc_grid_loc.begin(),loc.begin(),loc.end()); 
    proc_grid_dim.insert(proc_grid_dim.begin(),size.begin(),size.end());
    proc_grid_set = true; 
}

template <class T, class W> void ALL<T,W>::set_proc_grid_params(int* loc, int* size)
{
    proc_grid_loc.resize(balancer->get_dimension());
    proc_grid_dim.resize(balancer->get_dimension());
    for (int i = 0; i < balancer->get_dimension(); ++i)
    {
        proc_grid_loc.at(i) = loc[i];
        proc_grid_dim.at(i) = size[i];
    }
    proc_grid_set = true;
}

template <class T, class W> void ALL<T,W>::set_proc_tag(int tag)
{
    proc_tag = tag;
}

template <class T, class W> void ALL<T,W>::set_communicator(MPI_Comm comm)
{
    int comm_type;
    int topo_type;
    MPI_Topo_test(comm, &topo_type);
    if (topo_type == MPI_CART)
    {
        communicator = comm;
        balancer->set_communicator(communicator);
    }
    else
    {
        int rank;
        MPI_Comm_rank(comm, &rank);
        if (proc_grid_set)
        {
            // compute 1D index from the location in the process grid (using z-y-x ordering, as MPI_Dims_create)
            int idx;
            switch (balancer->get_dimension())
            {
                case 3:
                        idx = proc_grid_loc.at(2) +
                              proc_grid_loc.at(1) * proc_grid_dim.at(2) +
                              proc_grid_loc.at(0) * proc_grid_dim.at(2) * proc_grid_dim.at(1);
                        break;
                case 2:
                        idx = proc_grid_loc.at(1) +
                              proc_grid_loc.at(0) * proc_grid_dim.at(1);
                        break;
                case 1:
                        idx = proc_grid_loc.at(0);
                        break;
                default:
                    throw ALL_Internal_Error_Exception(
                                    __FILE__,
                                    __func__,
                                    __LINE__,
                                    "ALL cannot deal with more then three dimensions (or less then one).");
                    break;
            }

            // create new temporary communicator with correct ranks
            MPI_Comm temp_comm;
            MPI_Comm_split(comm, 0, idx, &temp_comm);

            std::vector<int> periods(3,1);

            // transform temporary communicator to cartesian communicator
            MPI_Cart_create(temp_comm, balancer->get_dimension(), proc_grid_dim.data(), periods.data(), 0, &communicator);
            balancer->set_communicator(communicator);
        }
        else
        {
            throw ALL_Internal_Error_Exception(
                            __FILE__,
                            __func__,
                            __LINE__,
                            "When using a non-cartesian communicator process grid parameters required (not set).");
        }
    }
}

template <class T, class W> void ALL<T,W>::calculate_outline()
{
    // calculation only possible if there are vertices defining the domain
    if (balancer->get_vertices().size() > 0)
    {
        outline.resize(2);
        // setup the outline with the first point
        for (int i = 0; i < balancer->get_dimension(); ++i)
        {
            outline.at(0) = balancer->get_vertices().at(0);
            outline.at(1) = balancer->get_vertices().at(0);
        }
        // compare value of each outline point with all vertices to find the maximum
        // extend of the domain in each dimension
        for (int i = 1; i < balancer->get_vertices().size(); ++i)
        {
            ALL_Point<T> p = balancer->get_vertices().at(i);
            for( int j = 0; j < balancer->get_dimension(); ++j)
            {
                outline.at(0).set_coordinate(j, std::min(outline.at(0).x(j),p.x(j)));
                outline.at(1).set_coordinate(j, std::max(outline.at(1).x(j),p.x(j)));
            }
        }
    }
}

template <class T, class W> void ALL<T,W>::get_work(W& result)
{
    result = balancer->get_work().at(0);
}

template <class T, class W> void ALL<T,W>::get_work(std::vector<W>& result)
{
    result = balancer->get_work();
}

template <class T, class W> void ALL<T,W>::setup()
{
    try
    {
        balancer->setup();
    }
    catch (ALL_Custom_Exception e)
    {
        std::cout << e.what() << std::endl;
        MPI_Abort(MPI_COMM_WORLD,-1);
    }
}

template <class T, class W> void ALL<T,W>::setup_balancer()
{
    try
    {
        balancer->setup(communicator);
    }
    catch (ALL_Custom_Exception e)
    {
        std::cout << e.what() << std::endl;
        MPI_Abort(MPI_COMM_WORLD,-1);
    }
}

template <class T, class W> std::vector<ALL_Point<T>>& ALL<T,W>::balance(bool internal)
{
    try
    {
        n_vertices = balancer->get_vertices().size();
        switch(method)
        {
            case ALL_LB_t::TENSOR:
                balancer->balance(loadbalancing_step);
                break;
            case ALL_LB_t::STAGGERED:
                
                /* ToDo: Check if repetition is useful at all and
                /*       change it to be included for all methods,
                    *       not only STAGGERED */
                /*
                // estimation to improve speed of convergence
                // changing vertices during estimation
                std::vector<ALL_Point<T>> tmp_vertices = balancer->get_vertices();
                // saved original vertices
                std::vector<ALL_Point<T>> old_vertices = balancer->get_vertices();
                // old temporary vertices
                std::vector<ALL_Point<T>> old_tmp_vertices = balancer->get_vertices();
                // result temporary vertices
                std::vector<ALL_Point<T>> result_vertices = balancer->get_vertices();
                // temporary work
                W tmp_work = balancer->get_work().at(0);
                // old temporary work
                W old_tmp_work = balancer->get_work().at(0);

                W max_work;
                W sum_work;
                double d_max;
                int n_ranks;

                // compute work density on local domain
                double V = 1.0;
                for (int i = 0; i < balancer->get_dimension(); ++i)
                    V *= ( outline.at(1).x(i) - outline.at(0).x(i) );
                double rho = work_array.at(0) / V;

                // collect maximum work in system
                MPI_Allreduce(work_array.data(), &max_work, 1, mpi_data_type_W, MPI_MAX, communicator);
                MPI_Allreduce(work_array.data(), &sum_work, 1, mpi_data_type_W, MPI_SUM, communicator);
                MPI_Comm_size(communicator,&n_ranks);
                d_max = (double)(max_work) * (double)n_ranks / (double)sum_work - 1.0;
                

                for (int i = 0; i < estimation_limit && d_max > 0.1 && !internal; ++i)
                {
                    balancer->set_work(tmp_work);
                    balancer->setup();
                    balancer->balance(true);
                    bool sane = true;
                    // check if the estimated boundaries are not too deformed
                    for (int j = 0; j < balancer->get_dimension(); ++j)
                        sane = sane && (result_vertices.at(0).x(j) <= old_vertices.at(1).x(j));
                    MPI_Allreduce(MPI_IN_PLACE,&sane,1,MPI_CXX_BOOL,MPI_LAND,communicator);
                    if (sane)
                    {
                        old_tmp_vertices = tmp_vertices;
                        tmp_vertices = result_vertices;
                        V = 1.0;
                        for (int i = 0; i < balancer->get_dimension(); ++i)
                            V *= ( tmp_vertices.at(1).x(i) - tmp_vertices.at(0).x(i) );
                        old_tmp_work = tmp_work;
                        tmp_work = rho * V;
                    }
                    else if (!sane || i == estimation_limit - 1)
                    {
                        balancer->get_vertices() = old_tmp_vertices;
                        work_array->at(0) = old_tmp_work;
                        calculate_outline();
                        i = estimation_limit;
                    }
                }

                ((ALL_Staggered_LB<T,W>*)balancer.get())->set_vertices(outline->data());
                */
                balancer->balance(loadbalancing_step);
                break;
            case ALL_LB_t::UNSTRUCTURED:
                balancer->balance(loadbalancing_step);
                break;
            case ALL_LB_t::VORONOI:
#ifdef ALL_VORONOI
                balancer->balance(loadbalancing_step);
#endif
                break;
            case ALL_LB_t::HISTOGRAM:
                balancer->balance(loadbalancing_step);
                break;
            default:
                throw ALL_Invalid_Argument_Exception(
                                        __FILE__,
                                        __func__,
                                        __LINE__,
                                        "Unknown type of loadbalancing passed.");
        }
        loadbalancing_step++;
    }
    catch(ALL_Custom_Exception e)
    {
        std::cout << e.what() << std::endl;
        MPI_Abort(MPI_COMM_WORLD,-1);
    }
    return balancer->get_vertices();

}

template <class T, class W> int ALL<T,W>::get_n_neighbors()
{
    std::vector<int> neig;
    try
    {
        balancer->get_neighbors(neig);
    }
    catch(ALL_Custom_Exception e)
    {
        std::cout << e.what() << std::endl;
        MPI_Abort(MPI_COMM_WORLD,-1);
    }
    return neig.size();
}

template <class T, class W> void ALL<T,W>::get_neighbor_vertices(std::vector<T>& nv)
{
    switch(method)
    {
        case ALL_LB_t::TENSOR:
            break;
        case ALL_LB_t::STAGGERED:
            break;
        case ALL_LB_t::UNSTRUCTURED:
            break;
        case ALL_LB_t::VORONOI:
#ifdef ALL_VORONOI
            ((ALL_Voronoi_LB<T,W>*)balancer.get())->get_neighbor_vertices(nv);
#endif
            break;
        case ALL_LB_t::HISTOGRAM:
            break;
        default:
            break;
    }
}

template <class T, class W> void ALL<T,W>::set_min_domain_size(const std::vector<T>& min_size)
{
    try
    {
        (balancer.get())->set_min_domain_size(min_size);
    }
    catch(ALL_Custom_Exception e)
    {
        std::cout << e.what() << std::endl;
        MPI_Abort(MPI_COMM_WORLD,-1);
    }
}

template <class T, class W> void ALL<T,W>::get_neighbors(std::vector<int>& neig)
{
    try
    {
        balancer->get_neighbors(neig);
    }
    catch(ALL_Custom_Exception e)
    {
        std::cout << e.what() << std::endl;
        MPI_Abort(MPI_COMM_WORLD,-1);
    }
}

template <class T, class W> void ALL<T,W>::get_neighbors(int** neig)
{
    try
    {
        balancer->get_neighbors(neig);
    }
    catch(ALL_Custom_Exception e)
    {
        std::cout << e.what() << std::endl;
        MPI_Abort(MPI_COMM_WORLD,-1);
    }
}

template <class T, class W> void ALL<T,W>::set_sys_size(std::vector<T>& s_size)
{
    balancer.get()->set_sys_size(s_size);
}

template <class T, class W> void ALL<T,W>::set_method_data(void* data)
{
    balancer.get()->set_data(data);
}

#ifdef ALL_VTK_OUTPUT
template <class T, class W> void ALL<T,W>::print_vtk_outlines(int step)
{
    // define grid points, i.e. vertices of local domain
    auto points = vtkSmartPointer<vtkPoints>::New();
    // allocate space for eight points (describing the cuboid around the domain)
    points->Allocate(8 * balancer->get_dimension());

    int n_ranks;
    int local_rank;

    if (!vtk_init)
    {    
        controller = vtkMPIController::New();
        controller->Initialize();
        controller->SetGlobalController(controller);
        vtk_init = true;
    }

    MPI_Comm_rank(communicator,&local_rank);
    MPI_Comm_size(communicator,&n_ranks);

    std::vector<ALL_Point<W>> tmp_outline(2);
    tmp_outline.at(0) = ALL_Point<W>(3,{outline.at(0).x(0),outline.at(0).x(1),outline.at(0).x(2)});
    tmp_outline.at(1) = ALL_Point<W>(3,{outline.at(1).x(0),outline.at(1).x(1),outline.at(1).x(2)});

    for (auto z = 0; z <= 1; ++z)
        for (auto y = 0; y <= 1; ++y)
            for (auto x = 0; x <= 1; ++x)
            {
                points->InsertPoint(x + 2 * y + 4 * z,
                                    tmp_outline.at(x).x(0),
                                    tmp_outline.at(y).x(1),
                                    tmp_outline.at(z).x(2));
            }

    auto hexa = vtkSmartPointer<vtkVoxel>::New();
    for (int i = 0; i < 8; ++i)
        hexa->GetPointIds()->SetId(i,i);

    auto cellArray = vtkSmartPointer<vtkCellArray>::New();
    cellArray->InsertNextCell(hexa);

    // define work array (length = 1)
    auto work = vtkSmartPointer<vtkFloatArray>::New();
    work->SetNumberOfComponents(1);
    work->SetNumberOfTuples(1);
    work->SetName("work");
    W total_work = std::accumulate(balancer->get_work().begin(),balancer->get_work().end(),(W)0);
    work->SetValue(0, total_work);

    // define rank array (length = 4, x,y,z, rank)
    int rank = 0;
    MPI_Comm_rank(communicator, &rank);
    auto rnk = vtkSmartPointer<vtkFloatArray>::New();
    rnk->SetNumberOfComponents(4);
    rnk->SetNumberOfTuples(1);
    rnk->SetName("rank");
    rnk->SetValue(0,proc_grid_loc.at(0));
    rnk->SetValue(1,proc_grid_loc.at(1));
    rnk->SetValue(2,proc_grid_loc.at(2));
    rnk->SetValue(3,rank);

    // define tag array (length = 1)
    auto tag = vtkSmartPointer<vtkIntArray>::New();
    tag->SetNumberOfComponents(1);
    tag->SetNumberOfTuples(1);
    tag->SetName("tag");
    tag->SetValue(0, proc_tag);


    // determine extent of system
    W global_extent[6];
    W local_min[3];
    W local_max[3];
    W global_min[3];
    W global_max[3];
    W width[3];
    W volume = (W)1.0;

    for(int i = 0; i < 3; ++i)
    {
        local_min[i] = outline.at(0).x(i);
        local_max[i] = outline.at(1).x(i);
        width[i] = local_max[i] - local_min[i];
        volume *= width[i];
    } 

    W surface = (W)2.0 * width[0] * width[1] +
                (W)2.0 * width[1] * width[2] +
                (W)2.0 * width[0] * width[2];

    auto expanse = vtkSmartPointer<vtkFloatArray>::New();
    expanse->SetNumberOfComponents(6);
    expanse->SetNumberOfTuples(1);
    expanse->SetName("expanse");
    expanse->SetValue(0, width[0]);
    expanse->SetValue(1, width[0]);
    expanse->SetValue(2, width[0]);
    expanse->SetValue(3, volume);
    expanse->SetValue(4, surface);
    expanse->SetValue(5, surface/volume);


    MPI_Allreduce(&local_min,&global_min,3,mpi_data_type_W,MPI_MIN,communicator);
    MPI_Allreduce(&local_max,&global_max,3,mpi_data_type_W,MPI_MAX,communicator);

    for(int i = 0; i < 3; ++i)
    {
        global_extent[2*i]   = global_min[i];
        global_extent[2*i+1] = global_max[i];
    }

    // create a structured grid and assign data to it
    auto unstructuredGrid = vtkSmartPointer<vtkUnstructuredGrid>::New();
    unstructuredGrid->SetPoints(points);
    unstructuredGrid->SetCells(VTK_VOXEL, cellArray);
    unstructuredGrid->GetCellData()->AddArray(work);
    unstructuredGrid->GetCellData()->AddArray(expanse);
    unstructuredGrid->GetCellData()->AddArray(rnk);
    unstructuredGrid->GetCellData()->AddArray(tag);

    

    std::ostringstream ss_local;
    ss_local << "vtk_outline/ALL_vtk_outline_" << std::setw(7) << std::setfill('0') 
             << step << "_" << local_rank << ".vtu";

    auto writer = vtkSmartPointer<vtkXMLUnstructuredGridWriter>::New();
    writer->SetInputData(unstructuredGrid);
    writer->SetFileName(ss_local.str().c_str());
    writer->SetDataModeToAscii();
    //writer->SetDataModeToBinary();
    writer->Write();

    //if (local_rank == 0)
    //{
        std::ostringstream ss_para;
        ss_para << "vtk_outline/ALL_vtk_outline_" << std::setw(7) << std::setfill('0') 
                << step << ".pvtu";
        // create the parallel writer
        auto parallel_writer = vtkSmartPointer<vtkXMLPUnstructuredGridWriter>::New();
        parallel_writer->SetFileName(ss_para.str().c_str());
        parallel_writer->SetNumberOfPieces(n_ranks);
        parallel_writer->SetStartPiece(local_rank);
        parallel_writer->SetEndPiece(local_rank);
        parallel_writer->SetInputData(unstructuredGrid);
        parallel_writer->SetDataModeToAscii();
        //parallel_writer->SetDataModeToBinary();
        parallel_writer->Write();
    //}
}

template <class T, class W> void ALL<T,W>::print_vtk_vertices(int step)
{
    try
    {
        int n_ranks;
        int local_rank;

        MPI_Comm_rank(communicator,&local_rank);
        MPI_Comm_size(communicator,&n_ranks);

        // local vertices
        // (vertices + work)
        T local_vertices[n_vertices * balancer->get_dimension() + 1];

        for (int v = 0; v < n_vertices; ++v)
        {
            for (int d = 0; d < balancer->get_dimension(); ++d)
            {
                local_vertices[v * balancer->get_dimension() + d] = balancer->get_vertices().at(v).x(d);
            }
        }
        local_vertices[n_vertices * balancer->get_dimension()] = (T)balancer->get_work().at(0);

        T* global_vertices;
        if (local_rank == 0)
        {
            global_vertices = new T[n_ranks * (n_vertices * balancer->get_dimension() + 1)];
        }

        // collect all works and vertices on a single process
        MPI_Gather(local_vertices, n_vertices * balancer->get_dimension() + 1, mpi_data_type_T,
                   global_vertices,n_vertices * balancer->get_dimension() + 1, mpi_data_type_T,
                   0, communicator
                  );

        if (local_rank == 0)
        {
            auto vtkpoints = vtkSmartPointer<vtkPoints>::New();
            auto unstructuredGrid = vtkSmartPointer<vtkUnstructuredGrid>::New();
    
            // enter vertices into unstructured grid
            for (int i = 0; i < n_ranks; ++i)
            {
                for (int v = 0; v < n_vertices; ++v)
                {
                    vtkpoints->InsertNextPoint(
                        global_vertices[ i * (n_vertices * balancer->get_dimension() + 1) + v * balancer->get_dimension() + 0 ],
                        global_vertices[ i * (n_vertices * balancer->get_dimension() + 1) + v * balancer->get_dimension() + 1 ],
                        global_vertices[ i * (n_vertices * balancer->get_dimension() + 1) + v * balancer->get_dimension() + 2 ]
                            );
                }
            }
            unstructuredGrid->SetPoints(vtkpoints);

            // data arrays for work and cell id
            auto work = vtkSmartPointer<vtkFloatArray>::New();
            auto cell = vtkSmartPointer<vtkFloatArray>::New();
            work->SetNumberOfComponents(1);
            work->SetNumberOfTuples(n_ranks);
            work->SetName("work");
            cell->SetNumberOfComponents(1);
            cell->SetNumberOfTuples(n_ranks);
            cell->SetName("cell id");

            for (int n = 0; n < n_ranks; ++n)
            {
                // define grid points, i.e. vertices of local domain
                vtkIdType pointIds[8] = { 8 * n + 0, 8 * n + 1, 8 * n + 2, 8 * n + 3, 
                                          8 * n + 4, 8 * n + 5, 8 * n + 6, 8 * n + 7 };

                auto faces = vtkSmartPointer<vtkCellArray>::New();
                // setup faces of polyhedron
                vtkIdType f0[3] = {8 * n + 0, 8 * n + 2, 8 * n + 1};
                vtkIdType f1[3] = {8 * n + 1, 8 * n + 2, 8 * n + 3};

                vtkIdType f2[3] = {8 * n + 0, 8 * n + 4, 8 * n + 2};
                vtkIdType f3[3] = {8 * n + 2, 8 * n + 4, 8 * n + 6};

                vtkIdType f4[3] = {8 * n + 2, 8 * n + 6, 8 * n + 3};
                vtkIdType f5[3] = {8 * n + 3, 8 * n + 6, 8 * n + 7};

                vtkIdType f6[3] = {8 * n + 1, 8 * n + 5, 8 * n + 3};
                vtkIdType f7[3] = {8 * n + 3, 8 * n + 5, 8 * n + 7};

                vtkIdType f8[3] = {8 * n + 0, 8 * n + 4, 8 * n + 1};
                vtkIdType f9[3] = {8 * n + 1, 8 * n + 4, 8 * n + 5};

                vtkIdType fa[3] = {8 * n + 4, 8 * n + 6, 8 * n + 5};
                vtkIdType fb[3] = {8 * n + 5, 8 * n + 6, 8 * n + 7};

                faces->InsertNextCell(3,f0);
                faces->InsertNextCell(3,f1);
                faces->InsertNextCell(3,f2);
                faces->InsertNextCell(3,f3);
                faces->InsertNextCell(3,f4);
                faces->InsertNextCell(3,f5);
                faces->InsertNextCell(3,f6);
                faces->InsertNextCell(3,f7);
                faces->InsertNextCell(3,f8);
                faces->InsertNextCell(3,f9);
                faces->InsertNextCell(3,fa);
                faces->InsertNextCell(3,fb);

                unstructuredGrid->InsertNextCell(VTK_POLYHEDRON, 8, pointIds,
                                                 12, faces->GetPointer());
                work->SetValue(n,global_vertices[ n * (n_vertices * balancer->get_dimension() + 1) + 
                                                  8 * balancer->get_dimension() ]);
                cell->SetValue(n,(T)n);
            }
            unstructuredGrid->GetCellData()->AddArray(work);
            unstructuredGrid->GetCellData()->AddArray(cell);

            std::ostringstream filename;
            filename << "vtk_vertices/ALL_vtk_vertices_" << 
                        std::setw(7) << std::setfill('0') << step << ".vtu";
            auto writer = vtkSmartPointer<vtkXMLUnstructuredGridWriter>::New();
            writer->SetInputData(unstructuredGrid);
            writer->SetFileName(filename.str().c_str());
            writer->SetDataModeToAscii();
            //writer->SetDataModeToBinary();
            writer->Write();

            delete[] global_vertices;
        }


        /*
        
        // define grid points, i.e. vertices of local domain
        vtkIdType pointIds[8] = { 0, 1, 2, 3, 4, 5, 6, 7 };

        auto points = vtkSmartPointer<vtkPoints>::New();

        int n_ranks;
        int local_rank;

        if (!vtk_init)
        {    
            controller = vtkMPIController::New();
            controller->Initialize();
            controller->SetGlobalController(controller_vertices);
            vtk_init = true;
        }

        MPI_Comm_rank(communicator,&local_rank);
        MPI_Comm_size(communicator,&n_ranks);

        // setup vertices of polyhedron
        if (n_vertices == 2)
        {
            for (auto z = 0; z <= 1; ++z)
                for (auto y = 0; y <= 1; ++y)
                    for (auto x = 0; x <= 1; ++x)
                    {
                        points->InsertPoint(x + 2 * y + 4 * z,
                                            vertices.at(x).x(0),
                                            vertices.at(y).x(1),
                                            vertices.at(z).x(2));
                    }
        }
        else if (n_vertices == 8)
        {
            for (int i = 0; i < 8; ++i)
            {
                points->InsertNextPoint(
                          vertices.at(i).x(0),
                          vertices.at(i).x(1),
                          vertices.at(i).x(2)
                                       );
            }
        }
        else
        {
            throw ALL_Invalid_Argument_Exception(   "ALL.hpp",
                                                    "ALL<T,W>::print_vtk_vertices(int)",
                                                    957,
                                                    "Unsupported number of vertices.");
        }

        // setup faces of polyhedron
        auto faces = vtkSmartPointer<vtkCellArray>::New();
        vtkIdType f0[3] = {0, 2, 1};
        vtkIdType f1[3] = {1, 2, 3};

        vtkIdType f2[3] = {0, 4, 2};
        vtkIdType f3[3] = {2, 4, 6};

        vtkIdType f4[3] = {2, 6, 3};
        vtkIdType f5[3] = {3, 6, 7};

        vtkIdType f6[3] = {1, 5, 3};
        vtkIdType f7[3] = {3, 5, 7};

        vtkIdType f8[3] = {0, 4, 1};
        vtkIdType f9[3] = {1, 4, 5};

        vtkIdType fa[3] = {4, 6, 5};
        vtkIdType fb[3] = {5, 6, 7};

        faces->InsertNextCell(3,f0);
        faces->InsertNextCell(3,f1);
        faces->InsertNextCell(3,f2);
        faces->InsertNextCell(3,f3);
        faces->InsertNextCell(3,f4);
        faces->InsertNextCell(3,f5);
        faces->InsertNextCell(3,f6);
        faces->InsertNextCell(3,f7);
        faces->InsertNextCell(3,f8);
        faces->InsertNextCell(3,f9);
        faces->InsertNextCell(3,fa);
        faces->InsertNextCell(3,fb);

        // define work array (length = 1)
        auto work = vtkSmartPointer<vtkFloatArray>::New();
        work->SetNumberOfComponents(1);
        work->SetNumberOfTuples(1);
        work->SetName("work");
        work->SetValue(0, work_array->at(0));

        // determine extent of system
        W global_extent[6];
        W local_min[3];
        W local_max[3];
        W global_min[3];
        W global_max[3];

        for(int i = 0; i < 3; ++i)
        {
            local_min[i] = outline->at(i);
            local_max[i] = outline->at(3+i);
        } 

        MPI_Allreduce(&local_min,&global_min,3,mpi_data_type_W,MPI_MIN,communicator);
        MPI_Allreduce(&local_max,&global_max,3,mpi_data_type_W,MPI_MAX,communicator);

        for(int i = 0; i < 3; ++i)
        {
            global_extent[2*i]   = global_min[i];
            global_extent[2*i+1] = global_max[i];
        }

        // create a structured grid and assign data to it
        auto unstructuredGrid = vtkSmartPointer<vtkUnstructuredGrid>::New();
        unstructuredGrid->SetPoints(points);
        unstructuredGrid->InsertNextCell(VTK_POLYHEDRON, 8, pointIds,
                          12, faces->GetPointer());
        unstructuredGrid->GetCellData()->AddArray(work);

        

        std::ostringstream ss_local;
        ss_local << "vtk_vertices/ALL_vtk_vertices_" << std::setw(7) << std::setfill('0') << step << "_" << local_rank << ".vtu";

        auto writer = vtkSmartPointer<vtkXMLUnstructuredGridWriter>::New();
        writer->SetInputData(unstructuredGrid);
        writer->SetFileName(ss_local.str().c_str());
        writer->SetDataModeToAscii();
        //writer->SetDataModeToBinary();
        writer->Write();

        //if (local_rank == 0)
        //{
            std::ostringstream ss_para;
            ss_para << "vtk_vertices/ALL_vtk_vertices_" << std::setw(7) << std::setfill('0') << step << ".pvtu";
            // create the parallel writer
            auto parallel_writer = vtkSmartPointer<vtkXMLPUnstructuredGridWriter>::New();
            parallel_writer->SetFileName(ss_para.str().c_str());
            parallel_writer->SetNumberOfPieces(n_ranks);
            parallel_writer->SetStartPiece(local_rank);
            parallel_writer->SetEndPiece(local_rank);
            parallel_writer->SetInputData(unstructuredGrid);
            parallel_writer->SetDataModeToAscii();
            //parallel_writer->SetDataModeToBinary();
            parallel_writer->Update();
            parallel_writer->Write();
        //}

        */            
    }
    catch (ALL_Custom_Exception e)
    {
        std::cout << e.what() << std::endl;
        MPI_Abort(MPI_COMM_WORLD,-1);
    }
}
#endif

#endif
