/*
Copyright 2018 Rene Halver, Forschungszentrum Juelich GmbH, Germany
Copyright 2018 Godehard Sutmann, Forschungszentrum Juelich GmbH, Germany

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this
   list of conditions and the following disclaimer in the documentation and/or
   other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may
   be used to endorse or promote products derived from this software without
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/


#ifndef ALL_LB_HEADER_INCLUDED
#define ALL_LB_HEADER_INCLUDED

#include <vector>
#include "ALL_Point.hpp"
#include "mpi.h"

// T: data type of vertices used in the balancer
// W: data type of the work used in the balancer
template <class T, class W> class ALL_LB {
 public:
    ALL_LB(int dim, T g) : 
        gamma(g),
        dimension(dim)
    {
        // set allowed minimum size to zero
        min_size = std::vector<T>(dim,0.0);
        periodicity.resize(dim);
        local_coords.resize(dim);
        global_dims.resize(dim);
    }
    virtual ~ALL_LB() = default;
    virtual void setup() = 0;
    virtual void balance(int step) = 0;
    // neighbors
    // provide list of neighbors
    virtual void get_neighbors(std::vector<int>&) = 0;
    // provide list of neighbors in each direction
    virtual void get_neighbors(int**) = 0;
    // setting / updating vertices
    virtual void set_vertices(std::vector<ALL_Point<T>>&);

    // global getter / setter functions
    virtual void set_dimension(const int d) {dimension = d;}
    virtual int get_dimension() {return dimension;}

    virtual void set_gamma(const T g) {gamma = g;}
    virtual T get_gamma() {return gamma;}

    virtual void set_work(const std::vector<W>& w) {work = w;}
    virtual void set_work(const W w) {work.resize(1); work.at(0) = w;}
    
    virtual void set_min_domain_size(const std::vector<T>& _min_size) 
        {min_size = _min_size;}
    virtual const std::vector<T>& get_min_domain_size() 
        {return min_size;}

    virtual void set_sys_size(const std::vector<T>& _sys_size) 
        {sys_size = _sys_size;}
    virtual const std::vector<T>& get_sys_size() 
        {return sys_size;}
        
    virtual std::vector<W>& get_work() {return work;}
    
    //virtual void set_vertices(std::vector<T>&);
    virtual std::vector<ALL_Point<T>>& get_vertices();
    virtual std::vector<ALL_Point<T>>& get_prev_vertices();
    
    virtual void set_communicator(MPI_Comm comm) 
    {
        // store communicator
        global_comm = comm;
        // get local rank within communicator
        MPI_Comm_rank(comm,&local_rank);
    }
    
    int get_n_vertices() {return vertices.size();}
    
    virtual void set_data(void*) = 0;
    
protected:
    // dimension of the system 
    int dimension;
    // correction factor / histogram bin size
    T gamma;
    // used communicator
    MPI_Comm global_comm;
    // local rank
    int local_rank;
    // if function supports a minimum domain size, use this
    std::vector<T> min_size;
    // system size
    std::vector<T> sys_size;
    // work array
    std::vector<W> work;
    // vertices
    std::vector<ALL_Point<T>> vertices;
    // previous (original) vertices
    std::vector<ALL_Point<T>> prev_vertices;
    // global dimension of the system in each coordinate
    std::vector<int> global_dims;
    // coordinates of the local domain in the system
    std::vector<int> local_coords;
    // periodicity of the MPI communicator / system
    std::vector<int> periodicity;
    // resize the vector to store the vertices
    void resize_vertices(int);
    
};

// set the actual vertices (unsafe due to array copy)
template <class T, class W> void ALL_LB<T,W>::set_vertices(std::vector<ALL_Point<T>>& vertices_in)
{
    // as a basic assumption the number of resulting vertices
    // is equal to the number of input vertices:
    // exceptions:
    //      VORONOI
    vertices = vertices_in;
    prev_vertices = vertices_in;
}

template <class T, class W> std::vector<ALL_Point<T>>& ALL_LB<T,W>::get_vertices()
{
    return vertices;
}

template <class T, class W> std::vector<ALL_Point<T>>& ALL_LB<T,W>::get_prev_vertices()
{
    return prev_vertices;
}

template <class T, class W> void ALL_LB<T,W>::resize_vertices(int new_size)
{
    vertices.resize(new_size);
}
#endif  // ALL_LB_HEADER_INCLUDED
