/*
Copyright 2018 Rene Halver, Forschungszentrum Juelich GmbH, Germany
Copyright 2018 Godehard Sutmann, Forschungszentrum Juelich GmbH, Germany

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may 
   be used to endorse or promote products derived from this software without 
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef ALL_TENSOR_HEADER_INCLUDED
#define ALL_TENSOR_HEADER_INCLUDED

#include <mpi.h>
#include <exception>
#include "ALL_CustomExceptions.hpp"
#include "ALL_LB.hpp"

// T: data type of vertices used in the balancer
// W: data type of the work used in the balancer
template <class T, class W> class ALL_Tensor_LB : public ALL_LB<T,W>
{
    public:
        ALL_Tensor_LB() {}
        ALL_Tensor_LB(int d, W w, T g) : 
            ALL_LB<T,W>(d, g)
        {
            this->set_work(w);
            // need the lower and upper bounds
            // of the domain for the tensor-based
            // load-balancing scheme

            // array of MPI communicators for each direction (to collect work
            // on each plane)
            communicators.resize(d);
            n_neighbors.resize(2*d);
        }

        ~ALL_Tensor_LB() = default;

        // setup communicators for the exchange of work in each domain
        void setup() override;

        // do a load-balancing step
        virtual void balance(int step) override;

        // getter for variables (passed by reference to avoid
        // direct access to private members of the object)

        // neighbors
        // provide list of neighbors
        void get_neighbors(std::vector<int>&) override;
        // provide list of neighbors in each direction
        void get_neighbors(int**) override;

        // empty override for setting method specific data
        virtual void set_data(void*) override {}
    private:
        // type for MPI communication
        MPI_Datatype mpi_data_type_T;
        MPI_Datatype mpi_data_type_W;

        // array of MPI communicators for each direction (to collect work
        // on each plane)
        std::vector<MPI_Comm> communicators;

        // list of neighbors
        std::vector<int> neighbors;
        std::vector<int> n_neighbors;
};

// setup routine for the tensor-based load-balancing scheme
// requires: 
//              this->global_comm (int): cartesian MPI communicator, from
//                                 which separate sub communicators
//                                 are derived in order to represent
//                                 each plane of domains in the system
template <class T, class W> void ALL_Tensor_LB<T,W>::setup()
{
    int status;

    // check if Communicator is cartesian
    MPI_Topo_test(this->global_comm, &status);
    if (status != MPI_CART)
    {
        throw ALL_Invalid_Comm_Type_Exception(
                __FILE__,
                __func__,
                __LINE__,
                "Cartesian MPI communicator required, passed communicator not cartesian");
    }

    int dim = this->get_dimension();
    
    // get the local coordinates, periodicity and global size from the MPI communicator
    MPI_Cart_get(this->global_comm, dim, this->global_dims.data(), this->periodicity.data(), this->local_coords.data());

    // get the local rank from the MPI communicator
    MPI_Cart_rank(this->global_comm, this->local_coords.data(), &this->local_rank);

    // create sub-communicators 
    for (int i = 0; i < dim; ++i)
    {
        MPI_Comm_split(this->global_comm,this->local_coords.at(i),0,&communicators[i]);
    }

    // determine correct MPI data type for template T
    if (std::is_same<T,double>::value) mpi_data_type_T = MPI_DOUBLE;
    else if (std::is_same<T,float>::value) mpi_data_type_T = MPI_FLOAT;
    else if (std::is_same<T,int>::value) mpi_data_type_T = MPI_INT;
    else if (std::is_same<T,long>::value) mpi_data_type_T = MPI_LONG;

    // determine correct MPI data type for template W 
    if (std::is_same<W,double>::value) mpi_data_type_W = MPI_DOUBLE;
    else if (std::is_same<W,float>::value) mpi_data_type_W = MPI_FLOAT;
    else if (std::is_same<W,int>::value) mpi_data_type_W = MPI_INT;
    else if (std::is_same<W,long>::value) mpi_data_type_W = MPI_LONG;

    // calculate neighbors
    int rank_left, rank_right;

    neighbors.clear();
    for (int i = 0; i < dim; ++i)
    {
        MPI_Cart_shift(this->global_comm,i,1,&rank_left,&rank_right);
        neighbors.push_back(rank_left);
        neighbors.push_back(rank_right);
        n_neighbors[2*i] = 1;
        n_neighbors[2*i+1] = 1;
    }
    
}

template<class T, class W> void ALL_Tensor_LB<T,W>::balance(int)
{
    int dim = this->get_dimension();
    std::vector<ALL_Point<T>>& vertices = this->get_vertices();
    std::vector<ALL_Point<T>>& prev_vertices = this->get_vertices();
    
    // loop over all available dimensions
    for (int i = 0; i < dim; ++i)
    {
        W work_local_plane;
        // collect work from all processes in the same plane
        MPI_Allreduce(this->get_work().data(),&work_local_plane,1,mpi_data_type_W,MPI_SUM,communicators.at(i));
        
        // correct right border:

        W remote_work;
        T local_size;
        T remote_size;
        // determine neighbors
        int rank_left, rank_right;

        MPI_Cart_shift(this->global_comm,i,1,&rank_left,&rank_right);

        // collect work from right neighbor plane
        MPI_Request sreq, rreq;
        MPI_Status sstat, rstat;

        MPI_Irecv(&remote_work,1,mpi_data_type_W,rank_right,0,this->global_comm,&rreq);
        MPI_Isend(&work_local_plane,1,mpi_data_type_W,rank_left,0,this->global_comm,&sreq);
        MPI_Wait(&sreq,&sstat);
        MPI_Wait(&rreq,&rstat);

        // collect size in dimension from right neighbor plane

        local_size = prev_vertices.at(1).x(i) - prev_vertices.at(0).x(i);
        
        MPI_Irecv(&remote_size,1,mpi_data_type_T,rank_right,0,this->global_comm,&rreq);
        MPI_Isend(&local_size,1,mpi_data_type_T,rank_left,0,this->global_comm,&sreq);
        MPI_Wait(&sreq,&sstat);
        MPI_Wait(&rreq,&rstat);

        // automatic selection of gamma to guarantee stability of method (choice of user gamma ignored)
        this->gamma = std::max(4.1, 1.1 * (1.0 + std::max(local_size,remote_size) / std::min(local_size,remote_size)));

        // calculate shift of borders:
        // s = 0.5 * gamma * (W_r - W_l) / (W_r + W_l) * (d_r + d_l)
        // *_r = * of neighbor (remote)
        // *_l = * of local domain
        T shift = (T)0;
        if (rank_right != MPI_PROC_NULL && this->local_coords.at(i) != this->global_dims.at(i)-1 && !( remote_work == (T)0 && work_local_plane == (T)0) )
            shift = 1.0 / this->gamma * 0.5 * (remote_work - work_local_plane) / (remote_work + work_local_plane) * (local_size + remote_size);

        // determine a maximum move in order to avoid overlapping borders and to guarantee
        // a stable shift of the domain borders (avoid the loss of a domain due to too large
        // shifts of its borders [no crossing of borders])
        T maxmove;
        if (shift > 0.0) 
            maxmove = 0.49*(remote_size-this->min_size[i]);
        else 
            maxmove = 0.49*(local_size-this->min_size[i]);

        if (std::fabs(shift) > maxmove)
        {
            shift = (shift < 0.0)?-maxmove:maxmove;
        }

        // send shift to right neighbors
        T remote_shift = (T)0;
        
        MPI_Irecv(&remote_shift,1,mpi_data_type_T,rank_left,0,this->global_comm,&rreq);
        MPI_Isend(&shift,1,mpi_data_type_T,rank_right,0,this->global_comm,&sreq);
        MPI_Wait(&sreq,&sstat);
        MPI_Wait(&rreq,&rstat);

        // if a left neighbor exists: shift left border 
        if (rank_left != MPI_PROC_NULL && this->local_coords[i] != 0)
            this->vertices.at(0).set_coordinate(i, prev_vertices.at(0).x(i) + remote_shift);
        else
            this->vertices.at(0).set_coordinate(i, prev_vertices.at(0).x(i));

        // if a right neighbor exists: shift right border
        if (rank_right != MPI_PROC_NULL && this->local_coords[i] != this->global_dims[i]-1)
            this->vertices.at(1).set_coordinate(i, prev_vertices.at(1).x(i) + shift);
        else
            this->vertices.at(1).set_coordinate(i, prev_vertices.at(1).x(i));
    }
}

// provide list of neighbors
template<class T, class W> void ALL_Tensor_LB<T,W>::get_neighbors(std::vector<int>& ret)
{
    ret = neighbors;
}
// provide list of neighbors in each direction
template<class T, class W> void ALL_Tensor_LB<T,W>::get_neighbors(int** ret)
{
    *ret = n_neighbors.data();
}
#endif
