/*
Copyright 2018 Rene Halver, Forschungszentrum Juelich GmbH, Germany
Copyright 2018 Godehard Sutmann, Forschungszentrum Juelich GmbH, Germany

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors may 
   be used to endorse or promote products derived from this software without 
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef ALL_UNSTRUCTURED_HEADER_INCLUDED
#define ALL_UNSTRUCTURED_HEADER_INCLUDED

/*

    Unstructured load-balancing scheme:

    Requirements: 
        a) cartesian communicator
        b) 2^(dim) vertices describing the domain (in theory could be more,
           we assume a orthogonal grid as start point)
        c) MPI >= 3.0 due to necessary non-blocking collectives


    Method:

        For each of the vertices of a domain a communicator is created, which contains the
        surrounding processes. For one of the vertices, a domain collects the centres of
        gravitiy and the work of the neighboring domains. In a next step a force acting on
        the vertex is computed as:
            F = 1.0 / ( gamma * n * sum(W_i) ) * sum( W_i * x_i )
            with:

                n:      number of neighboring domains
                W_i:    work on domain i
                x_i:    vector pointing from the vertex to 
                        the center of gravity of domain i
                gamma:  correction factor to control the
                        speed of the vertex shift
        
        */

#include <mpi.h>
#include <exception>
#include <vector>
#include <algorithm>
#include <cmath>
#include "ALL_CustomExceptions.hpp"
#include "ALL_Point.hpp"
#include "ALL_LB.hpp"

// T: data type of vertices used in the balancer
// W: data type of the work used in the balancer
template <class T, class W> class ALL_Unstructured_LB : public ALL_LB<T,W>
{
    public:
        ALL_Unstructured_LB() {}
        ALL_Unstructured_LB(int d, W w, T g) : 
            ALL_LB<T,W>(d, g)
        {
            this->set_work(w);
            // need the lower and upper bounds
            // of the domain for the tensor-based
            // load-balancing scheme
            vertex_rank.resize(8);
        }

        ~ALL_Unstructured_LB() override;

        // setup communicators for the exchange of work in each domain
        void setup() override;

        // do a load-balancing step
        void balance(int step) override;

        // getter for variables (passed by reference to avoid
        // direct access to private members of the object)

        // neighbors
        // provide list of neighbors
        void get_neighbors(std::vector<int>&) override;
        // provide list of neighbors in each direction
        void get_neighbors(int**) override;

        // empty override for setting method specific data
        virtual void set_data(void*) override {}
        
    private:
        // MPI values

        // rank(s) in the communicators of the vertices
        std::vector<int> vertex_rank;

        // type for MPI communication
        MPI_Datatype mpi_data_type_T;
        MPI_Datatype mpi_data_type_W;

        // array of MPI communicators for each direction (to collect work
        // on each plane)
        std::vector<MPI_Comm> communicators;

        // list of neighbors
        std::vector<int> neighbors;

        // find neighbors (more sophisticated than for tensor-product
        // variant of LB)
        void find_neighbors();
        
        // number of vertices (since it is not equal for all domains)
        int n_vertices;
};

template <class T, class W> ALL_Unstructured_LB<T,W>::~ALL_Unstructured_LB()
{
}

// provide list of neighbors
template<class T, class W> void ALL_Unstructured_LB<T,W>::get_neighbors(std::vector<int>& ret)
{
    ret = neighbors;
}
// provide list of neighbors in each direction
template<class T, class W> void ALL_Unstructured_LB<T,W>::get_neighbors(int** ret)
{
    *ret = neighbors.data();
}

template <class T, class W> void ALL_Unstructured_LB<T,W>::setup()
{
    n_vertices = this->vertices.size();
    communicators = std::vector<MPI_Comm>(n_vertices);
    
    // determine correct MPI data type for template T
    if (std::is_same<T,double>::value) mpi_data_type_T = MPI_DOUBLE;
    else if (std::is_same<T,float>::value) mpi_data_type_T = MPI_FLOAT;
    else if (std::is_same<T,int>::value) mpi_data_type_T = MPI_INT;
    else if (std::is_same<T,long>::value) mpi_data_type_T = MPI_LONG;
    else
    {
        throw ALL_Invalid_Comm_Type_Exception(
                __FILE__,
                __func__,
                __LINE__,
                "Invalid data type for boundaries given (T)");
    }

    // determine correct MPI data type for template W 
    if (std::is_same<W,double>::value) mpi_data_type_W = MPI_DOUBLE;
    else if (std::is_same<W,float>::value) mpi_data_type_W = MPI_FLOAT;
    else if (std::is_same<W,int>::value) mpi_data_type_W = MPI_INT;
    else if (std::is_same<W,long>::value) mpi_data_type_W = MPI_LONG;
    else
    {
        throw ALL_Invalid_Comm_Type_Exception(
                __FILE__,
                __func__,
                __LINE__,
                "Invalid data type for work given (W)");
    }

    MPI_Group all;
    MPI_Comm_group(this->global_comm,&all);

    // check if communicator is cartesian
    int status;
    MPI_Topo_test(this->global_comm, &status);

    if (status != MPI_CART)
    {
        throw ALL_Invalid_Comm_Type_Exception(
                __FILE__,
                __func__,
                __LINE__,
                "Cartesian MPI communicator required, passed communicator is not cartesian");
    }

    // get the local coordinates, this->periodicity and global size from the MPI communicator
    MPI_Cart_get(this->global_comm, this->dimension, this->global_dims.data(), this->periodicity.data(), this->local_coords.data());

    // get the local rank from the MPI communicator
    MPI_Cart_rank(this->global_comm, this->local_coords.data(), &this->local_rank);

    // groups required for new communicators
    MPI_Group groups[n_vertices];
    // arrays of processes belonging to group
    int processes[n_vertices][n_vertices];

    // shifted local coordinates to find neighboring processes
    int shifted_coords[this->dimension];

    /*

    // setup communicators for each of the vertices
    if (dimension == 1)
    {

        // sequence of vertices in 1d
        //
        // 0 - - - 1

        for (int x = 0; x < 2; ++x)
        {
            // group / communicator for vertex 0
            shifted_coords[0] = this->local_coords.at(0) - 1 + x;

            MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[x][0])); 

            shifted_coords[0] = this->local_coords.at(0) + x;

            MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[x][1])); 

            MPI_Group_incl(all, 2, &processes[x][0], &groups[x]);
            MPI_Comm_create(this->global_comm, groups[x], &communicators[x]);

            // get local rank in vertex communicator
            MPI_Comm_rank(communicators[x],&vertex_rank[x]);
        }
    }            
    else if (dimension == 2)
    {
        // sequence of vertices in 2d
        //
        // 2 - - - 3
        // |       |
        // |       |
        // 0 - - - 1

        for (int y = 0; y < 2; ++y)
        {
            for (int x = 0; x < 2; ++x)
            {
                int vertex = 2 * y + x;

                shifted_coords[0] = this->local_coords.at(0) - 1 + x;
                shifted_coords[1] = this->local_coords.at(1) - 1 + y;

                MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][0])); 

                shifted_coords[0] = this->local_coords.at(0) + x;
                shifted_coords[1] = this->local_coords.at(1) - 1 + y;

                MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][1])); 

                shifted_coords[0] = this->local_coords.at(0) - 1 + x;
                shifted_coords[1] = this->local_coords.at(1) + y;

                MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][2])); 

                shifted_coords[0] = this->local_coords.at(0) + x;
                shifted_coords[1] = this->local_coords.at(1) + y;

                MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][3])); 

                MPI_Group_incl(all, 4, &processes[vertex][0], &groups[vertex]);
                MPI_Comm_create(this->global_comm, groups[vertex], &communicators[vertex]);

                // get local rank in vertex communicator
                MPI_Comm_rank(communicators[vertex],&vertex_rank[vertex]);
            }
        }
    }
    else if (dimension == 3)
    {
        // sequence of vertices in 3d
        //
        //    6 - - - 7
        //   /|      /|
        //  / |     / |
        // 4 -2- - 5  3
        // | /     | /
        // |/      |/
        // 0 - - - 1

        for (int z = 0; z < 2; ++z)
        {
            for (int y = 0; y < 2; ++y)
            {
                for (int x = 0; x < 2; ++x)
                {
                    int vertex = 4 * z + 2 * y + x;

                    // group / communicator for vertex 0,2,4,6
                    shifted_coords[0] = this->local_coords.at(0) - 1 + x;
                    shifted_coords[1] = this->local_coords.at(1) - 1 + y;
                    shifted_coords[2] = this->local_coords.at(2) - 1 + z;

                    MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][0])); 

                    shifted_coords[0] = this->local_coords.at(0) + x;
                    shifted_coords[1] = this->local_coords.at(1) - 1 + y;
                    shifted_coords[2] = this->local_coords.at(2) - 1 + z;

                    MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][1])); 

                    shifted_coords[0] = this->local_coords.at(0) - 1 + x;
                    shifted_coords[1] = this->local_coords.at(1) + y;
                    shifted_coords[2] = this->local_coords.at(2) - 1 + z;

                    MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][2])); 

                    shifted_coords[0] = this->local_coords.at(0) + x;
                    shifted_coords[1] = this->local_coords.at(1) + y;
                    shifted_coords[2] = this->local_coords.at(2) - 1 + z;

                    MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][3])); 

                    shifted_coords[0] = this->local_coords.at(0) - 1 + x;
                    shifted_coords[1] = this->local_coords.at(1) - 1 + y;
                    shifted_coords[2] = this->local_coords.at(2) + z;

                    MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][4])); 

                    shifted_coords[0] = this->local_coords.at(0) + x;
                    shifted_coords[1] = this->local_coords.at(1) - 1 + y;
                    shifted_coords[2] = this->local_coords.at(2) + z;

                    MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][5])); 

                    shifted_coords[0] = this->local_coords.at(0) - 1 + x;
                    shifted_coords[1] = this->local_coords.at(1) + y;
                    shifted_coords[2] = this->local_coords.at(2) + z;

                    MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][6])); 

                    shifted_coords[0] = this->local_coords.at(0) + x;
                    shifted_coords[1] = this->local_coords.at(1) + y;
                    shifted_coords[2] = this->local_coords.at(2) + z;

                    MPI_Cart_rank(this->global_comm, shifted_coords, &(processes[vertex][7])); 

                    MPI_Group_incl(all, 8, &processes[vertex][0], &groups[vertex]);
                    MPI_Comm_create(this->global_comm, groups[vertex], &communicators[vertex]);

                    // get local rank in vertex communicator
                    MPI_Comm_rank(communicators[vertex],&vertex_rank[vertex]);
                }
            }
        }
        for (int i = 0; i < 8; ++i)
        {
            if (i == this->local_rank)
            {
                std::cout << "local rank: " << i << " " << std::endl;
                for (int k = 0; k < 8; ++k)
                {
                    std::cout << "group " << k << ": ";
                    for (int j = 0; j < 8; ++j)
                        std::cout << processes[k][j] << " ";
                    std::cout << std::endl;
                }
                std::cout << std::endl;
                MPI_Barrier(MPI_COMM_WORLD);
            }
            else
                MPI_Barrier(MPI_COMM_WORLD);

        }
    }
    else
    {
        throw ALL_Invalid_Argument_Exception(   "ALL_Unstructured.hpp",
                                                "ALL_Unstructured_LB<T,W>::setup(int)",
                                                257,
                                                "Unsupported number of dimensions.");
    }
    */


#ifdef ALL_DEBUG_ENABLED
    MPI_Barrier(this->global_comm);
    if (this->local_rank == 0) std::cout << "ALL_Unstructured<T,W>::setup() preparing communicators..." << std::endl;    
#endif    
    std::vector<int> dim_vert(this->global_dims);

    MPI_Comm tmp_comm;
    int own_vertex;

    for (int i = 0; i < n_vertices; ++i)
        communicators.at(i) = MPI_COMM_SELF;

#ifdef ALL_DEBUG_ENABLED
    MPI_Barrier(this->global_comm);
    if (this->local_rank == 0) std::cout << "ALL_Unstructured<T,W>::setup() computing communicators..." << std::endl;
    std::cout << "DEBUG: " << " rank: " << this->local_rank
                           << " dim_vert: " << dim_vert.at(0) << " " << dim_vert.at(1) << " " << dim_vert.at(2)
                           << " size(local_coords): " << this->local_coords.size() << " "
                           << " size(global_dims):  " << this->global_dims.size()
                           << " size(vertex_rank):  " << vertex_rank.size()
                           << std::endl;
#endif    
    for (int iz = 0; iz < dim_vert.at(2); ++iz)
    {
        for (int iy = 0; iy < dim_vert.at(1); ++iy)
        {
            for (int ix = 0; ix < dim_vert.at(0); ++ix)
            {   
                if (ix == ( ( this->local_coords.at(0) + 0 ) % this->global_dims.at(0))  &&
                    iy == ( ( this->local_coords.at(1) + 0 ) % this->global_dims.at(1))  &&
                    iz == ( ( this->local_coords.at(2) + 0 ) % this->global_dims.at(2)) )
                    own_vertex = 0;
                else if (ix == ( ( this->local_coords.at(0) + 1 ) % this->global_dims.at(0))  &&
                         iy == ( ( this->local_coords.at(1) + 0 ) % this->global_dims.at(1))  &&
                         iz == ( ( this->local_coords.at(2) + 0 ) % this->global_dims.at(2)) )
                    own_vertex = 1;
                else if (ix == ( ( this->local_coords.at(0) + 0 ) % this->global_dims.at(0))  &&
                         iy == ( ( this->local_coords.at(1) + 1 ) % this->global_dims.at(1))  &&
                         iz == ( ( this->local_coords.at(2) + 0 ) % this->global_dims.at(2)) )
                    own_vertex = 2;
                else if (ix == ( ( this->local_coords.at(0) + 1 ) % this->global_dims.at(0))  &&
                         iy == ( ( this->local_coords.at(1) + 1 ) % this->global_dims.at(1))  &&
                         iz == ( ( this->local_coords.at(2) + 0 ) % this->global_dims.at(2)) )
                    own_vertex = 3;
                else if (ix == ( ( this->local_coords.at(0) + 0 ) % this->global_dims.at(0))  &&
                         iy == ( ( this->local_coords.at(1) + 0 ) % this->global_dims.at(1))  &&
                         iz == ( ( this->local_coords.at(2) + 1 ) % this->global_dims.at(2)) )
                    own_vertex = 4;
                else if (ix == ( ( this->local_coords.at(0) + 1 ) % this->global_dims.at(0))  &&
                         iy == ( ( this->local_coords.at(1) + 0 ) % this->global_dims.at(1))  &&
                         iz == ( ( this->local_coords.at(2) + 1 ) % this->global_dims.at(2)) )
                    own_vertex = 5;
                else if (ix == ( ( this->local_coords.at(0) + 0 ) % this->global_dims.at(0))  &&
                         iy == ( ( this->local_coords.at(1) + 1 ) % this->global_dims.at(1))  &&
                         iz == ( ( this->local_coords.at(2) + 1 ) % this->global_dims.at(2)) )
                    own_vertex = 6;
                else if (ix == ( ( this->local_coords.at(0) + 1 ) % this->global_dims.at(0))  &&
                         iy == ( ( this->local_coords.at(1) + 1 ) % this->global_dims.at(1))  &&
                         iz == ( ( this->local_coords.at(2) + 1 ) % this->global_dims.at(2)) )
                    own_vertex = 7;
                else 
                    own_vertex = -1;

                // if vertex belongs to local domain, join communicator, otherwise ignore it
                int result;
                if (own_vertex >= 0)
                {
                    vertex_rank.at(own_vertex) = 7 - own_vertex;
                    result = MPI_Comm_split(this->global_comm,1,vertex_rank.at(own_vertex),&communicators.at(own_vertex));
                    MPI_Comm_rank(communicators.at(own_vertex),&vertex_rank.at(own_vertex));
                }
                else
                {
                    result = MPI_Comm_split(this->global_comm,0,0,&tmp_comm);
                }
                if (result != MPI_SUCCESS)
                {
                    std::cout << "RANK: " << this->local_rank << " ERROR in MPI_Comm_Split" << std::endl;
                }
            }
        }
#ifdef ALL_DEBUG_ENABLED
    MPI_Barrier(this->global_comm);
    if (this->local_rank == 0) std::cout << "ALL_Unstructured<T,W>::setup() finished computing communicators..." << std::endl;    
#endif    
    }
    for (int vertex = 0; vertex < n_vertices; ++vertex)
        MPI_Comm_rank(communicators.at(vertex),&vertex_rank.at(vertex));

}

// TODO: periodic boundary conditions (would require size of the system)
//       for now: do not change the vertices along the edge of the system

template <class T, class W> void ALL_Unstructured_LB<T,W>::balance(int /*step*/)
{

    this->prev_vertices = this->vertices;

    // geometrical center and work of each neighbor for each vertex
    // work is cast to vertex data type, since in the end a shift
    // of the vertex is computed
    T vertex_info[n_vertices*n_vertices*(this->dimension+2)];
    T center[this->dimension];

    // local geometric center
    T local_info[(this->dimension+2) * n_vertices];

    for (int i = 0; i < n_vertices*n_vertices*(this->dimension+2); ++i)
        vertex_info[i] = -1;

    for (int d = 0; d < (this->dimension+2) * n_vertices; ++d)
        local_info[d] = (T)0;

    for (int d = 0; d < this->dimension; ++d)
        center[d] = (T)0;

    // compute geometric center
    for (int v = 0; v < n_vertices; ++v)
    {
        for (int d = 0; d < this->dimension; ++d)
        {
            center[d] += ( (this->prev_vertices.at(v)).x(d) / (T)n_vertices );
        }
        local_info[v * (this->dimension+2) + this->dimension] = (T)this->work.at(0);
    }
    for (int v = 0; v < n_vertices; ++v)
    {
        for (int d = 0; d < this->dimension; ++d)
        {
            local_info[v * (this->dimension+2) + d] = center[d] - (this->prev_vertices.at(v)).x(d);
        }
    }

    // compute maximum distance of a vertex to the plane of the neighboring three vertices
    auto distance = [=] ( std::vector<ALL_Point<T>>* verts, int p, int a, int b, int c, int rank )
    {
        // vectors spanning the plane from vertex 'a'
        T vb[3];
        T vc[3];
        
        for (int i = 0; i < 3; ++i)
        {
            vb[i] = verts->at(b).x(i) - verts->at(a).x(i);
            vc[i] = verts->at(c).x(i) - verts->at(a).x(i);
        }
        
        // normal vector of plane
        T n[3];

        n[0] = vb[1] * vc[2] - vb[2] * vc[1];        
        n[1] = vb[2] * vc[0] - vb[0] * vc[2];        
        n[2] = vb[0] * vc[1] - vb[1] * vc[0];

        // length of normal vector
        T dn = sqrt
            (
                n[0] * n[0] +
                n[1] * n[1] +
                n[2] * n[2]
            );

        // distance from origin
        T d = ( n[0] * verts->at(a).x(0) +
                n[1] * verts->at(a).x(1) +
                n[2] * verts->at(a).x(2)); 

        /*
        if (rank == 4)
        {
            std::cout   << p << " " << std::endl
                        << " A: " << verts->at(a).x(0) << " " << verts->at(a).x(1) << " " << verts->at(a).x(2) << " " << std::endl
                        << " B: " << verts->at(b).x(0) << " " << verts->at(b).x(1) << " " << verts->at(b).x(2) << " " << std::endl
                        << " C: " << verts->at(c).x(0) << " " << verts->at(c).x(1) << " " << verts->at(c).x(2) << " " << std::endl
                        << " b: " << vb[0] << " " << vb[1] << " " << vb[2] << std::endl
                        << " c: " << vc[0] << " " << vc[1] << " " << vc[2] << std::endl
                        << " n: " << n[0] << " " << n[1] << " " << n[2] << std::endl
                        << " |n|: " << dn << std::endl
                        << " d: " << d << std::endl
                        << " dp: " << std::abs (
                                                (
                                                   n[0] * verts->at(p).x(0)
                                                   + n[1] * verts->at(p).x(1)
                                                   + n[2] * verts->at(p).x(2)
                                                   - d
                                                ) / dn
                                              ) << std::endl;
        }
        */

        // compute distance of test vertex to plane
        return std::abs (
                    (
                          n[0] * verts->at(p).x(0) 
                        + n[1] * verts->at(p).x(1)
                        + n[2] * verts->at(p).x(2)
                        - d
                    ) / dn
               );

    };

    // compute for each vertex the maximum movement towards the center

    local_info[ 0 * (this->dimension+2) + this->dimension+1 ] =
        distance( &(this->prev_vertices), 0, 1, 2, 4, this->local_rank);

    local_info[ 1 * (this->dimension+2) + this->dimension+1 ] =
        distance( &(this->prev_vertices), 1, 0, 3, 5, this->local_rank);

    local_info[ 2 * (this->dimension+2) + this->dimension+1 ] =
        distance( &(this->prev_vertices), 2, 3, 0, 6, this->local_rank);

    local_info[ 3 * (this->dimension+2) + this->dimension+1 ] =
        distance( &(this->prev_vertices), 3, 2, 1, 7, this->local_rank);

    local_info[ 4 * (this->dimension+2) + this->dimension+1 ] =
        distance( &(this->prev_vertices), 4, 5, 6, 0, this->local_rank);

    local_info[ 5 * (this->dimension+2) + this->dimension+1 ] =
        distance( &(this->prev_vertices), 5, 4, 7, 1, this->local_rank);

    local_info[ 6 * (this->dimension+2) + this->dimension+1 ] =
        distance( &(this->prev_vertices), 6, 7, 4, 2, this->local_rank);

    local_info[ 7 * (this->dimension+2) + this->dimension+1 ] =
        distance( &(this->prev_vertices), 7, 6, 5, 3, this->local_rank);

    /*
    if (this->local_rank == 4)
    {
        std::cout << "max dists: " 
                  << local_info[ 0 * (dimension+2) + dimension+1 ] << " "
                  << local_info[ 1 * (dimension+2) + dimension+1 ] << " "
                  << local_info[ 2 * (dimension+2) + dimension+1 ] << " "
                  << local_info[ 3 * (dimension+2) + dimension+1 ] << " "
                  << local_info[ 4 * (dimension+2) + dimension+1 ] << " "
                  << local_info[ 5 * (dimension+2) + dimension+1 ] << " "
                  << local_info[ 6 * (dimension+2) + dimension+1 ] << " "
                  << local_info[ 7 * (dimension+2) + dimension+1 ] << " "
                  << work << std::endl;
    }
    */


    // exchange information with all vertex neighbors
    MPI_Request request[n_vertices];
    MPI_Status status[n_vertices];

    for (int v = 0; v < n_vertices; ++v)
    {
        MPI_Iallgather(&local_info[v*(this->dimension+2)],this->dimension+2,mpi_data_type_T,
                       &vertex_info[v*n_vertices*(this->dimension+2)],this->dimension+2,mpi_data_type_T,
                       communicators[v],&request[v]);
    }
    MPI_Waitall(n_vertices,request,status);

    // compute new position for vertex 7 (if not periodic)
    T total_work = (T)0;
    T shift_vectors[this->dimension * n_vertices];

    for (int v = 0; v < n_vertices; ++v)
    {
        for (int d = 0; d < this->dimension; ++d)
        {
            shift_vectors[v * this->dimension + d] = (T)0;
        }
    }

    // get total work
    for (int v = 0; v < n_vertices; ++v)
    {
        if (vertex_info[(n_vertices-1)*n_vertices+v*(this->dimension+2)+this->dimension] > 1e-6)
            total_work += vertex_info[(n_vertices-1)*n_vertices+v*(this->dimension+2)+this->dimension];
    }

    // determine shortest distance between vertex and a neighbor 
    T cdist = std::max( 
                        1e-6,
                        vertex_info[(n_vertices-1)*(n_vertices*(this->dimension+2))+0*(this->dimension+2)+this->dimension+1]
                      );

    for (int v = 1; v < n_vertices; ++v)
    {
        if ( vertex_info[(n_vertices-1)*(n_vertices*(this->dimension+2))+v*(this->dimension+2)+this->dimension+1] > 0 )
        {
            cdist = std::min
                    (
                        cdist,
                        vertex_info[(n_vertices-1)*(n_vertices*(this->dimension+2))+v*(this->dimension+2)+this->dimension+1]
                    );
        }
    }
    cdist *= 0.49;

    cdist = 0.5;

    if (std::abs(total_work) >= 1e-6)
    {
        for (int i = 0; i < n_vertices; ++i)
        {
            for (int d = 0; d < this->dimension; ++d)
            {
                if (this->local_coords.at(d) != this->global_dims.at(d) - 1)
                {
                    //if ( vertex_info[(n_vertices-1)*(n_vertices*(dimension+2))+i*(dimension+2)+dimension] - ( total_work / (T)n_vertices )  > 0)                    
                    //{
                        shift_vectors[(n_vertices-1)*this->dimension + d] += 
                            ( vertex_info[(n_vertices-1)*(n_vertices*(this->dimension+2))+i*(this->dimension+2)+this->dimension] - ( total_work / (T)n_vertices ) ) *
                            vertex_info[(n_vertices-1)*(n_vertices*(this->dimension+2))+i*(this->dimension+2)+d]  
                            / ( (T)2.0 * this->gamma * total_work * (T)n_vertices);
                    //}
                }
            }
        }
        // distance of shift
        T shift = (T)0;
        for (int d = 0; d < this->dimension; ++d)
            shift +=  shift_vectors[(n_vertices-1)*this->dimension + d] * shift_vectors[(n_vertices-1)*this->dimension + d];
        shift = sqrt(shift);

        if (std::abs(shift) >= 1e-6 && shift > cdist)
        {
            // normalize length of shift vector to half the distance to the nearest center
            for (int d = 0; d < this->dimension; ++d)
                shift_vectors[(n_vertices-1)*this->dimension + d] *= cdist / shift;
        }

    }

    for (int i = 0; i < n_vertices; ++i)
    {
        MPI_Ibcast(&shift_vectors[i*this->dimension],this->dimension,mpi_data_type_T,0,communicators[i],&request[i]);
    }
    MPI_Waitall(n_vertices,request,status);

    /*
    this->vertices = this->prev_vertices;
    if (this->local_rank == 7)
    {
        std::cout << "center: ";
        for (int d = 0; d < dimension; ++d)
        {
            std::cout << center[d] << " ";
        }
        std::cout << std::endl;
    }
    */
    for (int v = 0; v < n_vertices; ++v)
    {
        //if (this->local_rank == 4) std::cout << "vertex " << v << ": ";
        for (int d = 0; d < this->dimension; ++d)
        {
            this->vertices.at(v).set_coordinate(d, this->prev_vertices.at(v).x(d) + shift_vectors[v * this->dimension + d]);
            if (this->vertices.at(v).x(d) != this->vertices.at(v).x(d))
            {
                std::cout << this->local_rank << " " << v << " " << d << std::endl;
                MPI_Abort(MPI_COMM_WORLD,-200);
            }
            //if (this->local_rank == 4) std::cout << shifted_vertices->at(v).x(d) << " " << shift_vectors[v * dimension + d] << " ";
        }
        //if (this->local_rank == 4) std::cout << std::endl;
    }
}

#endif

