! -DCMAKE_Fortran_FLAGS='-std=f2003 -Wall -Wextra -fbacktrace' -DCM_ALL_FORTRAN=ON -DCMAKE_BUILD_TYPE=Debug -DCM_ALL_DEBUG=ON
module ALL_module_obj
    use iso_c_binding
    use iso_fortran_env
    implicit none
    private
    ! The different loadbalancing methods. Must be kept in sync with C side.
    ! TODO reuse those from C side (is that possible?)
    integer(c_int), public, parameter :: ALL_STAGGERED = 0
    integer(c_int), public, parameter :: ALL_TENSOR = 1
    integer(c_int), public, parameter :: ALL_UNSTRUCTURED = 2
    integer(c_int), public, parameter :: ALL_VORONOI = 3
    integer(c_int), public, parameter :: ALL_HISTOGRAM = 4
    ! Direct interface with C wrapper
    ! TODO put this into separate file, and only have the wrappers and convenience functions here
    !      The interface should not be public! Only this module counts as the public API!
    !      We do not consider the C functions public! Can we somehow keep them private?
    ! TODO support optionally mpi_f08 using CMAKE_ALL_USE_F08 / ALL_USE_F08
    ! TODO add intent() to dummy arguments
    interface
        function all_init_f(method, dim, gamma) result(this) bind(C)
            use iso_c_binding
            integer(c_int), value :: method
            integer(c_int), value :: dim
            real(c_double), value :: gamma
            type(c_ptr) :: this
        end function
        subroutine all_set_proc_grid_params_f(obj, nloc, loc, nsize, size) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
            integer(c_int), value :: nloc
            integer(c_int), dimension(nloc) :: loc
            integer(c_int), value :: nsize
            integer(c_int), dimension(nsize) :: size
        end subroutine
        subroutine all_set_proc_tag_f(obj, tag) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
            integer(c_int), value :: tag
        end subroutine
        subroutine all_set_min_domain_size_f(obj, dim, domain_size) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
            integer(c_int), value :: dim
            real(c_double), dimension(dim) :: domain_size
        end subroutine
        subroutine all_set_work_f(obj, work) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
            real(c_double), value :: work
        end subroutine
        ! TODO is dim here really necessary? How should the public API be, just vertices I assume
        !      check the C++ API
        subroutine all_set_vertices_f(obj, n, dim, vertices) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
            integer(c_int), value :: n
            integer(c_int), value :: dim
            real(c_double), dimension(n*dim) :: vertices
        end subroutine
        ! TODO Change for mpi_f08!
        subroutine all_set_communicator_f(obj, comm) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
            integer, value :: comm
        end subroutine
        subroutine all_setup_f(obj) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
        end subroutine
        subroutine all_balance_f(obj) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
        end subroutine
        subroutine all_get_new_number_of_vertices_f(obj, n) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
            integer(c_int) :: n
        end subroutine
        ! TODO is n here needed? may it be inout?
        ! TODO automatically get number of vertices
        ! TODO check length of vertices (need dim for that, length is n*dim)
        subroutine all_get_vertices_f(obj, n, vertices) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
            integer(c_int), value :: n
            real(c_double), dimension(*) :: vertices
        end subroutine
        ! TODO same as all_get_vertices_f
        subroutine all_get_prev_vertices_f(obj, n, vertices) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
            integer(c_int), value :: n
            real(c_double), dimension(*) :: vertices
        end subroutine
        subroutine all_print_vtk_outlines_f(obj, step) bind(C)
            use iso_c_binding
            type(c_ptr), value :: obj
            integer(c_int), value :: step
        end subroutine
    end interface
    type, public :: ALL_t
        type(c_ptr), private :: object = C_NULL_PTR !< pointer to C++ object used on C side
        integer, private :: dim = 0 !< dimensionality of system, set during init
    contains
        procedure :: init => ALL_init
        procedure :: set_proc_grid_params => ALL_set_proc_grid_params
        procedure :: set_proc_tag => ALL_set_proc_tag
        procedure :: set_min_domain_size => ALL_set_min_domain_size
        procedure :: set_work => ALL_set_work
        procedure :: set_vertices => ALL_set_vertices
        procedure :: set_communicator => ALL_set_communicator
        procedure :: setup => ALL_setup
        procedure :: balance => ALL_balance
        procedure :: get_new_number_of_vertices => ALL_get_new_number_of_vertices
        procedure :: get_vertices => ALL_get_vertices
        procedure :: get_vertices_alloc => ALL_get_vertices_alloc
        procedure :: get_prev_vertices => ALL_get_prev_vertices
        procedure :: print_vtk_outlines => ALL_print_vtk_outlines
    end type
contains
    !> Initializes the loadbalancing object
    subroutine ALL_init(this, method, dim, gamma)
        class(ALL_t), intent(out) :: this
        integer, intent(in) :: method !< Must be one of the ALL_* method values
        integer, intent(in) :: dim !< dimensionality of system
        real(c_double), intent(in) :: gamma !< gamma value for load balancer (ignored for TENSOR and STAGGERED)
        this%object = all_init_f(method, dim, gamma)
        this%dim = dim
    end subroutine
    subroutine ALL_set_proc_grid_params(this, loc, ranks)
        class(ALL_t), intent(in) :: this
        integer, dimension(this%dim), intent(in) :: loc !< index of domain in `dim` directions (0-indexed!)
        integer, dimension(this%dim), intent(in) :: ranks !< total number of domains in `dim` directions
        ! TODO these nloc and nsize, must always be dim? Which has been set previously!
        !      If the lengths are irrelevant, we may do away with the safety and just bind the C function to
        !      the derived type.
        call all_set_proc_grid_params_f(this%object, this%dim, loc, this%dim, ranks)
    end subroutine
    subroutine ALL_set_proc_tag(this, tag)
        class(ALL_t), intent(in) :: this
        integer, intent(in) :: tag !< tag of local process, only output in VTK outlines
        call all_set_proc_tag_f(this%object, tag)
    end subroutine
    subroutine ALL_set_min_domain_size(this, domain_size)
        class(ALL_t), intent(in) :: this
        real(c_double), dimension(this%dim), intent(in) :: domain_size !< minimum domain size
        call all_set_min_domain_size_f(this%object, this%dim, domain_size)
    end subroutine
    subroutine ALL_set_work(this, work)
        class(ALL_t), intent(in) :: this
        real(c_double), intent(in) :: work !< work of this domain
        call all_set_work_f(this%object, work)
    end subroutine
    subroutine ALL_set_vertices(this, vertices)
        class(ALL_t), intent(in) :: this
        real(c_double), dimension(:,:), intent(in) :: vertices !< vertices of domain, for `n` domains must have shape: (dim,n)
#ifndef NDEBUG
        if(size(vertices,1) /= this%dim) then
            write(error_unit,'(a)') "ALL_set_vertices: Wrong dimension for vertices"
            stop
        endif
#endif
        call all_set_vertices_f(this%object, size(vertices,2), this%dim, vertices)
    end subroutine
    subroutine ALL_set_communicator(this, comm)
#ifdef ALL_USE_F08
        use mpi_f08
#endif
        class(ALL_t), intent(in) :: this
#ifndef ALL_USE_F08
        integer, intent(in) :: comm !< MPI Communicator, not type checked!
        call all_set_communicator_f(this%object, comm)
#else
        type(MPI_Comm), intent(in) :: comm !< MPI Communicator
        call all_set_communicator_f(this%object, comm%mpi_val)
#endif
    end subroutine
    subroutine ALL_setup(this)
        class(ALL_t), intent(in) :: this
        call all_setup_f(this%object)
    end subroutine
    subroutine ALL_balance(this)
        class(ALL_t), intent(in) :: this
        call all_balance_f(this%object)
    end subroutine
    subroutine ALL_get_new_number_of_vertices(this, n)
        class(ALL_t), intent(in) :: this
        integer(c_int), intent(out) :: n !< set to number of new vertices
        call all_get_new_number_of_vertices_f(this%object, n)
    end subroutine
    subroutine ALL_get_vertices(this, vertices)
        class(ALL_t), intent(in) :: this
        real(c_double), dimension(:,:), intent(out) :: vertices !< set to new vertices, must be exact size (dim,n)
        integer(c_int) :: n_verts
#ifndef NDEBUG
        call this%get_new_number_of_vertices(n_verts)
        if(size(vertices,1) /= this%dim .or. size(vertices,2) < n_verts) then
            write(error_unit,'(a)') "ALL_get_vertices: vertices array not large enough!"
            stop
        endif
#endif
        call all_get_vertices_f(this%object, size(vertices,2), vertices)
    end subroutine
    !> Same function as get_vertices, but takes an allocatable array, and resizes automatically
    subroutine ALL_get_vertices_alloc(this, vertices)
        class(ALL_t), intent(in) :: this
        real(c_double), dimension(:,:), allocatable, intent(inout) :: vertices !< set to new vertices, may be reallocated to fit
        integer(c_int) :: n_verts
        call this%get_new_number_of_vertices(n_verts)
        if(allocated(vertices)) then
            if(size(vertices,1) /= this%dim .or. size(vertices,2) < n_verts) then
                deallocate(vertices)
            endif
        endif
        if(.not.allocated(vertices)) allocate(vertices(this%dim, n_verts))
        call all_get_vertices_f(this%object, size(vertices,2), vertices)
    end subroutine
    subroutine ALL_get_prev_vertices(this, vertices)
        class(ALL_t), intent(in) :: this
        real(c_double), dimension(:,:), intent(out) :: vertices !< set to prev vertices, must be exact size (dim,n), unchecked!
#ifndef NDEBUG
        ! TODO Check against number of vertices not only dimensionality
        if(size(vertices,1) /= this%dim) then
            write(error_unit,'(a)') "ALL_get_vertices: vertices array not large enough!"
            stop
        endif
#endif
        call all_get_prev_vertices_f(this%object, size(vertices,2), vertices)
    end subroutine
    subroutine ALL_print_vtk_outlines(this, step)
        class(ALL_t), intent(in) :: this
        integer(c_int), intent(in) :: step !< current step, used for filename
        call all_print_vtk_outlines_f(this%object, step)
    end subroutine
end module
